/*
	Sudoku generator

	Motif interface

	(C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>
#include <Xm/Separator.h>
#include <Xm/ToggleB.h>
#include <Xm/Protocols.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>
#include <Xm/CutPaste.h>
#include <Xm/Separator.h>

static Widget main_wnd=NULL, printer_wnd=NULL, settings_wnd=NULL;
static Widget toplevel;
static Widget main_form, settings_pb, copy_pb, paste_pb, create_pb, solve_pb, print_pb, quit_pb, drawing_wnd;
static Widget main_settingsdialog, settings_form, settings_label, settings_edit,
	      separator1, separator2,
              settings_radio2, symmetrical1, symmetrical2, symmetrical3,
	      settings_radio, level1, level2, level3, level4, level5,
              settings_ok, settings_cancel;
static Widget main_printdialog, printer_fileform, printer_fileselection, papersize_label, papersize,
              papersize_pane, A4_w, Letter_w, printer_explain_label, printer_explain_check,
	      printer_explain_select;

typedef struct
{
  XFontStruct *textfont;
  Pixel redcolor, blackcolor;
} ApplicationData, *ApplicationDataPtr;

static ApplicationData applicationdata;

static XtResource resources[] =
{
  {"textfont", "Textfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(ApplicationDataPtr, textfont),
   XtRString, "-ibm-courier-medium-r-normal--50-*-*-*-m-*-iso8859-1"
  },
  {"redcolor", "Redcolor", XtRPixel, sizeof(Pixel),
   XtOffset(ApplicationDataPtr, redcolor),
   XtRString, "red"
  },
  {"blackcolor", "Blackcolor", XtRPixel, sizeof(Pixel),
   XtOffset(ApplicationDataPtr, blackcolor),
   XtRString, "black"
  },
};

static GC thick_gc, thin_gc, black_gc, red_gc;
static int font_width,font_height;

int nfilled=28;
int newnfilled;
int bestndigits=81;
int ndigits;
int symmetrical=1;
int level=1;
static int newlevel=1;
static int newsymmetric=1;
static char filled_buffer[4];

#define THINWIDTH 1
#define THICKWIDTH 3
#define CELLSIZE 50


int displayboard[9][9];

int setboard[9][9];
int setboardsave[9][9];
int beforeboard[9][9];
int board[9][9];
int nmove=0;
int nsol=0;
int ntest=0;
int running=0;
int letter=0;
int print_explain=1;
int solved=0;
int solution_unique=0;

int n;
Arg wargs[1000];

int nlines=0;
char *loglines=NULL;
int logsize=0;

void loginit()
{
  if(loglines!=NULL) free(loglines);
  loglines=NULL;
  logsize=0;
  nlines=0;
}

void logend()
{
}

void logprintf(const char *template, ...)
{
  va_list ap;
  char tmpbuf[1024];
  int len,i;

  va_start(ap, template);
  vsprintf(tmpbuf, template, ap);
  va_end(ap);
  len=strlen(tmpbuf);
  for(i=0;i<len;i++)
  {
    if(tmpbuf[i]=='\n') nlines++;
  }
  
  if(loglines==NULL)
  {
    loglines=malloc(len+1);
    loglines[0] = '\0';
  }
  else
  {
    loglines=realloc(loglines, strlen(loglines)+len+1);
  }
  strcat(loglines, tmpbuf);
}

setcreatemenutext(char *txt)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(txt))); n++;
  XtSetValues(create_pb, wargs, n);

  XmStringFree(stmp);
}

setsolvemenutext(char *txt)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(txt))); n++;
  XtSetValues(solve_pb, wargs, n);

  XmStringFree(stmp);
}

settitle(char *txt)
{
  XTextProperty text_prop;

  text_prop.value = (unsigned char *) txt;
  text_prop.encoding = XA_STRING;
  text_prop.format = 8;
  text_prop.nitems = strlen(txt);
  XSetWMName(XtDisplay(toplevel), XtWindow(toplevel), &text_prop);
}

static void generate_sudoku()
{
  bestndigits=81;
  running=1;
  ntest=0;
}

static printer_print()
{
}

static main_clear()
{
  int i,j;
  for(i=0;i<9;i++)
  for(j=0;j<9;j++)
  displayboard[i][j] = 0;
}

main_update()
{
  char txt[16];
  int row,col,size,x,y,width;

  size = 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH;

  y=0;
  XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), thin_gc,
	    0, 0, size, 0);
  y+=THINWIDTH;
  for(row=0;row<9;row++)
  {
    if(row>0)
    {
      XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), (row%3==0)?thick_gc:thin_gc,
                y, 0, y, size);
      XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), (row%3==0)?thick_gc:thin_gc,
                0, y, size, y);
      y+=(row%3==0)?THICKWIDTH:THINWIDTH;
    }
    x=0;
    for(col=0;col<9;col++)
    {
      if(displayboard[row][col])
      {
        sprintf(txt, "%d", abs(displayboard[row][col]));
      }
      else
      {
        txt[0]=' ';
	txt[1]=0;
      }
      XDrawImageString(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), displayboard[row][col]<0?red_gc:black_gc, x+CELLSIZE/2-font_width/2, y+CELLSIZE-(CELLSIZE-applicationdata.textfont->ascent)/2, txt, strlen(txt));
      if(col>0) x+=(col%3==0)?THICKWIDTH:THINWIDTH;
      x+=CELLSIZE;
    }
    y+=CELLSIZE;
  }
}

static void settings_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(!running)
  {
    newlevel=level;
    newsymmetric=symmetrical;
    XtRealizeWidget(main_settingsdialog);
  }
}

static void copy_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *textboard=" . . . | . . . | . . . \n\
 . . . | . . . | . . . \n\
 . . . | . . . | . . . \n\
-------+-------+------ \n\
 . . . | . . . | . . . \n\
 . . . | . . . | . . . \n\
 . . . | . . . | . . . \n\
-------+-------+------ \n\
 . . . | . . . | . . . \n\
 . . . | . . . | . . . \n\
 . . . | . . . | . . . \n";
  int i,r,c;
  char buf[512];
  long item_id = 0;
  int status;
  XmString clip_label;
  static int cnt=0;
  Display *dpy = XtDisplayOfObject(w);
  Window window = XtWindowOfObject(w);

  sprintf(buf, "Sudoku-%d", ++cnt);
  clip_label = XmStringCreateLocalized("Sudoku");

  r=0;
  c=0;
  i=0;
  for(i=0; i<=strlen(textboard); i++)
  {
    if(textboard[i]=='.')
    {
      if(displayboard[r][c]>0)
      {
	buf[i] = '0'+displayboard[r][c];
      }
      else
      {
	buf[i] = '.';
      }
      c++;
      if(c==9)
      {
	c=0;
	r++;
      }
    }
    else
    {
      buf[i] = textboard[i];
    }
  }

  do
    status = XmClipboardStartCopy(dpy, window, clip_label, CurrentTime, NULL, NULL, &item_id);
  while(status == ClipboardLocked);

  XmStringFree(clip_label);

  do
    status = XmClipboardCopy(dpy, window, item_id, "STRING",
             buf, (long) strlen(buf)+1, cnt, NULL);
  while(status == ClipboardLocked);

  do
    status = XmClipboardEndCopy(dpy, window, item_id);
  while(status==ClipboardLocked);
}

static void paste_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  int status;
  unsigned long received;
  char *data, buf[1024];
  Display *dpy = XtDisplayOfObject (w);
  Window window = XtWindowOfObject (w);
  int i,r,c,formatok;
  char ch;
  int newboard[9][9];
  if(!running)
  {
    if(solved)
    {
      unsolve_sudoku();
      main_update();
      setsolvemenutext("Solve");
    }
    do
      status = XmClipboardStartRetrieve (dpy, window, CurrentTime);
    while (status == ClipboardLocked);
    status = XmClipboardRetrieve (dpy, window, "STRING", buf, sizeof(buf)-1, &received, NULL);
    buf[received] = 0; /* NULL terminate */
    XmClipboardEndRetrieve (dpy, window);
    if (status == ClipboardSuccess)
    {
      formatok=1;
      i=0;
      r=0;
      c=0;
      while(ch=buf[i++])
      {
	if(ch=='.' || (ch>='0' && ch<='9'))
	{
	  if(ch=='.') ch='0';
	  if(r>=9)
	  {
	    formatok=0;
	  }
	  else
	  {
	    newboard[r][c] = ch-'0';
	    c++;
	    if(c==9)
	    {
	      c=0;
	      r++;
	    }
	  }
	}
      }
      if(r!=9 || c!=0) formatok=0;
      for(r=0;r<9;r++)
      for(c=0;c<9;c++)
      displayboard[r][c] = formatok?newboard[r][c]:0;
      main_update();
    }
  }
}

static void create_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char windowtitle[80];
  if(running)
  {
    setcreatemenutext("Create");
    running=0;
    sprintf(windowtitle, "Sudoku Generator. Stopped at iteration %d, best: %d", ntest, bestndigits);
    settitle(windowtitle);
  }
  else
  {
    setcreatemenutext("Stop!");
    generate_sudoku();
    setsolvemenutext("Solve");
   }
}

static int OKPressed, Wnd_destroyed;

static void ok_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  OKPressed = TRUE;
} 

static void destroy_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  Wnd_destroyed = TRUE;
}

void informuser(Widget refwnd,char *txt)
{
  Widget inform_shell, form;
  Widget stext, ok_pb;
  int x, y;
  XmString stmp;

  OKPressed = FALSE;
  Wnd_destroyed = FALSE;

  /* Create a popup shell for the information window */
  n = 0;
  XtSetArg(wargs[n], XtNtitle, "Inform"); n++;
  XtSetArg(wargs[n], XmNmessageString, (stmp=XmStringCreateLocalized(txt))); n++;
  XtSetArg(wargs[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg(wargs[n], XmNautoUnmanage, FALSE); n++;
  XtSetArg(wargs[n], XmNnoResize, TRUE); n++;
  XtSetArg(wargs[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  XtSetArg(wargs[n], XmNdeleteResponse, XmDESTROY); n++;
  inform_shell = XmCreateInformationDialog(refwnd, "inform", wargs, n);
  XmStringFree(stmp);

  XtUnmanageChild(XmMessageBoxGetChild(inform_shell, XmDIALOG_CANCEL_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(inform_shell, XmDIALOG_HELP_BUTTON));
  				
  XtAddCallback(inform_shell, XmNdestroyCallback, destroy_callback, NULL);
  XtAddCallback(inform_shell, XmNokCallback, ok_callback, NULL);

  XtManageChild(inform_shell);

  /* Local event loop - wait for ok */
  while (!OKPressed)
  {
    XEvent event;

    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }

  /* Popdown and destroy the information window */
  if (!Wnd_destroyed)
  {
    XtUnmanageChild(inform_shell);
    XtDestroyWidget(inform_shell);
  }
}

static void solve_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(!running)
  {
   if(solved)
   {
     unsolve_sudoku();
     main_update();
     setsolvemenutext("Solve");
   }
   else
   {
     solution_unique=solve_sudoku(0);
     if(solution_unique==0)
     {
       informuser(toplevel, "This sudoku does not have a solution.");
     }
     else if(solution_unique>1)
     {
       informuser(toplevel, "This sudoku has more than one solution.");
     }
     main_update();
     if(solved) setsolvemenutext("Unsolve");
   }
  }
}

static void print_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionDoSearch(printer_fileselection, NULL);
  XtRealizeWidget(main_printdialog);
}

static void quit_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  exit(0);
}

static void drawing_expose_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  main_update();
}

static void level_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  newlevel = (int) client_data;
}

static void symmetrical_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  newsymmetric = (int) client_data;
}

static void settingsok_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *s;

  s = XmTextFieldGetString(settings_edit);
  sscanf(s, "%d", &nfilled);
  XFree(s);

  symmetrical=newsymmetric;
  level=newlevel;

  main_clear();
  bestndigits=81;
  settitle("Sudoku Generator");
  main_update();

  XtUnrealizeWidget(main_settingsdialog);
}

static void settingscancel_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnrealizeWidget(main_settingsdialog);
}

static void printer_fileopen_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call_data;
  char *filename;
  FILE *print_fh;
  double dev_xmin, dev_xmax, dev_ymin, dev_ymax, margin, yspacing;
  char *paper_text;
  double size, printer_cellsize, x, y, fontsize, width, printer_thinwidth, printer_thickwidth;
  int row, col, i;

  XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);

  n=0;
  print_explain=0;
  XtSetArg(wargs[n], XmNset, &print_explain); n++;
  XtGetValues(printer_explain_select, wargs, n);

  print_fh = fopen(filename, "w");
  if(print_fh != NULL)
  {
    if(letter)
    {
      dev_xmin = 0;
      dev_xmax = 612;
      dev_ymin = 0;
      dev_ymax = 792;
      margin = 18;
      paper_text = "Letter";
    }
    else
    {
      dev_xmin = 0;
      dev_xmax = 595;
      dev_ymin = 0;
      dev_ymax = 842;
      margin = 18;
      paper_text = "A4";
    }
    
    fprintf(print_fh, "%%!PS-Adobe-3.0\n");
    fprintf(print_fh, "%%%%Creator: Sudoku generator\n");
    fprintf(print_fh, "%%%%Pages: (atend)\n");
    fprintf(print_fh, "%%%%DocumentMedia: %s %.0f %.0f 0 white ()\n", paper_text, dev_xmax, dev_ymax);
    fprintf(print_fh, "%%%%BoundingBox: %.0f %.0f %.0f %.0f\n", dev_xmin, dev_ymin, dev_xmax, dev_ymax);
    fprintf(print_fh, "%%%%Orientation: Portrait\n");
    fprintf(print_fh, "%%%%EndComments\n\n");

    fprintf(print_fh, "[{\n");
    fprintf(print_fh, "%%%%BeginFeature: *PageSize %s\n", paper_text);
    fprintf(print_fh, "<</PageSize[%d %d]/ImagingBBox null>>setpagedevice\n", (int) dev_xmax, (int) dev_ymax);
    fprintf(print_fh, "%%%%EndFeature\n");
    fprintf(print_fh, "} stopped cleartomark\n");

    if((dev_xmax-dev_xmin)>(dev_ymax-dev_ymin))
    {
      size = dev_ymax-dev_ymin;
    }
    else
    {
      size = dev_xmax-dev_xmin;
    }

    size = size-2*margin;

    if(solved && print_explain && solution_unique==1)
    {
      size = size/5;
      fprintf(print_fh, "gsave %g %g translate\n", (dev_xmax-dev_xmin)/2-size/2, dev_ymax-margin-size);
    }
    else
    {
      fprintf(print_fh, "%g %g translate\n", (dev_xmax-dev_xmin)/2-size/2, (dev_ymax-dev_ymin)/2-size/2);
    }

    fontsize = size/10;
    printer_cellsize=size/9;
    printer_thinwidth = size/400;
    printer_thickwidth = printer_thinwidth*3;

    fprintf(print_fh, "/showcenter {dup gsave 0 0 moveto true charpath flattenpath pathbbox grestore /symax exch def /sxmax exch def /symin exch def /sxmin exch def sxmin sxmax sub 2 div sxmin sub symin symax sub 2 div rmoveto show} def\n");
    fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);

    y=size;
    for(row=0;row<9;row++)
    {
      if(row>0)
      {
        width = (row%3==0)?printer_thickwidth:printer_thinwidth;
	fprintf(print_fh, "%g setlinewidth\n", width);
	fprintf(print_fh, "0 0 0 setrgbcolor\n");
	fprintf(print_fh, "newpath\n");
	fprintf(print_fh, "%g 0 moveto\n", y);
	fprintf(print_fh, "%g %g lineto\n", y, size);
	fprintf(print_fh, "0 %g moveto\n", y);
	fprintf(print_fh, "%g %g lineto\n", size, y);
	fprintf(print_fh, "stroke\n");
      }
      x=0;
      for(col=0; col<9; col++)
      {
	if(displayboard[row][col])
	{
	  fprintf(print_fh, "%g %g moveto\n", x+printer_cellsize/2,y-printer_cellsize/2);
	  fprintf(print_fh, "%d 0 0 setrgbcolor\n", displayboard[row][col]<0?1:0);
	  fprintf(print_fh, "(%d) showcenter\n", abs(displayboard[row][col]));
	}
	x+=printer_cellsize;
      }
      y-=printer_cellsize;
    }
    fprintf(print_fh, "%g setlinewidth\n", printer_thickwidth);
    fprintf(print_fh, "0 0 0 setrgbcolor\n");
    fprintf(print_fh, "newpath\n");
    fprintf(print_fh, "0 0 moveto\n");
    fprintf(print_fh, "0 %g lineto\n", size);
    fprintf(print_fh, "%g %g lineto\n", size, size);
    fprintf(print_fh, "%g 0 lineto\n", size);
    fprintf(print_fh, "closepath\n");
    fprintf(print_fh, "stroke\n");
    if(solved && print_explain)
    {
      /* Print out explanation: */
      unsolve_sudoku();
      loginit();
      solve_sudoku(1);
      logend();
      yspacing = (dev_ymax-size-dev_ymin-2*margin)/(nlines+1);
      if(yspacing>15) yspacing=15;
      fontsize = yspacing/1.25;
      fprintf(print_fh, "grestore %g %g translate\n", dev_xmin+margin, dev_ymax-margin-yspacing-size);
      fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);
      y=0;
      fprintf(print_fh, "0 0 moveto (");
      for(i=0;i<strlen(loglines);i++)
      {
        if(loglines[i]=='\n')
	{
	  fprintf(print_fh, ") show\n");
	  y -= yspacing;
	  fprintf(print_fh, "0 %g moveto (",y);
	}
	else if(loglines[i]=='(' ||
	        loglines[i]==')')
	{
	  fprintf(print_fh, "\\%c", loglines[i]);
	}
	else
	{
	  fprintf(print_fh, "%c", loglines[i]);
	}
      }
      fprintf(print_fh, ") show\n");
    }
    fprintf(print_fh, "showpage\n");
    fprintf(print_fh, "%%%%Pages: 1\n");
    fprintf(print_fh, "%%%%EOF\n%c", (char) 4);

    fclose(print_fh);
  }

  XtFree(filename);
  XtUnrealizeWidget(main_printdialog);
}

static void printer_fileclose_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnrealizeWidget(main_printdialog);
}

static void printer_papersize_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  letter = (int) client_data;
}

static main_wnd_init()
{
  Atom atom;
  Widget menubar;
  XGCValues values;
  XtGCMask valueMask;

  XmString s[20];
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XtNtitle, "Sudoku"); n++;
  XtSetArg(wargs[n], XtNgeometry, "+10+10"); n++;
  main_wnd = XmCreateMainWindow(toplevel, "main", wargs, n);
  atom = XmInternAtom(XtDisplay(toplevel),"WM_DELETE_WINDOW",TRUE);
  XmAddWMProtocolCallback(toplevel, atom, quit_cb, False);
  n=0;
  main_form = XtCreateManagedWidget("main_form", xmFormWidgetClass,
                                    main_wnd, wargs, n);

  /* Menu bar */

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  settings_pb = XtCreateManagedWidget("Settings", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(settings_pb, XmNactivateCallback, settings_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_pb); n++;
  copy_pb = XtCreateManagedWidget("Copy", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(copy_pb, XmNactivateCallback, copy_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, copy_pb); n++;
  paste_pb = XtCreateManagedWidget("Paste", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(paste_pb, XmNactivateCallback, paste_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, paste_pb); n++;
  create_pb = XtCreateManagedWidget("Create", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(create_pb, XmNactivateCallback, create_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, create_pb); n++;
  solve_pb = XtCreateManagedWidget("Solve", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(solve_pb, XmNactivateCallback, solve_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, solve_pb); n++;
  print_pb = XtCreateManagedWidget("Print", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(print_pb, XmNactivateCallback, print_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, print_pb); n++;
  quit_pb = XtCreateManagedWidget("Quit", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(quit_pb, XmNactivateCallback, quit_cb, NULL);

  /* Main drawing area */

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_pb); n++;
  XtSetArg(wargs[n], XmNwidth, 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH); n++;
  XtSetArg(wargs[n], XmNheight, 9*CELLSIZE+2*THICKWIDTH+7*THINWIDTH); n++;
  drawing_wnd = XtCreateManagedWidget("drawing_wnd", xmDrawingAreaWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(drawing_wnd, XmNexposeCallback, drawing_expose_cb, (XtPointer) NULL);

  
  valueMask = GCForeground | GCBackground | GCLineWidth;
  values.foreground = BlackPixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.line_width = THINWIDTH;
  thin_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCLineWidth;
  values.foreground = BlackPixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.line_width = THICKWIDTH;
  thick_gc = XtGetGC(drawing_wnd, valueMask, &values);

  font_width=applicationdata.textfont->max_bounds.width;
  font_height=applicationdata.textfont->ascent+applicationdata.textfont->descent;

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.blackcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  black_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.redcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  red_gc = XtGetGC(drawing_wnd, valueMask, &values);

  n=0;
  XtSetArg(wargs[n], XtNtitle, "Sudoku settings"); n++;
  main_settingsdialog = XtCreateApplicationShell("sudoku_settings", topLevelShellWidgetClass, wargs, n);

  n=0;
  settings_form = XtCreateManagedWidget("form", xmFormWidgetClass, main_settingsdialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 8); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("No. of filled cells:"))); n++;
  settings_label = XmCreateLabel(settings_form, "FilledTitle", wargs, n);
  XtManageChild(settings_label);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_label); n++;
  XtSetArg(wargs[n], XmNvalue, filled_buffer); n++;
  sprintf(filled_buffer, "%d", nfilled);
  XtSetArg(wargs[n], XmNmaxLength, sizeof(filled_buffer)-1); n++;
  XtSetArg(wargs[n], XmNcolumns, sizeof(filled_buffer)-1); n++;
  XtSetArg(wargs[n], XmNverifyBell, False); n++;

  settings_edit = XtCreateManagedWidget("settings_edit", xmTextFieldWidgetClass, settings_form, wargs, n);
  XtManageChild(settings_edit);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_edit); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  separator1 = XtCreateManagedWidget("separator1", xmSeparatorWidgetClass, settings_form, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, separator1); n++;
  settings_radio2 = XmCreateRadioBox(settings_form, "RadioBox2", wargs, n);
  XtManageChild(settings_radio2);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Not symmetric"))); n++;
  XtSetArg(wargs[n], XmNset, (symmetrical==0)?True:False); n++;
  symmetrical1 = XmCreateToggleButton(settings_radio2, "ToggleSimple21", wargs, n);
  XtManageChild(symmetrical1);
  XmStringFree(s[0]);
  XtAddCallback(symmetrical1, XmNarmCallback, symmetrical_cb, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("X&Y mirror"))); n++;
  XtSetArg(wargs[n], XmNset, (symmetrical==1)?True:False); n++;
  symmetrical2 = XmCreateToggleButton(settings_radio2, "ToggleSimple22", wargs, n);
  XtManageChild(symmetrical2);
  XmStringFree(s[0]);
  XtAddCallback(symmetrical2, XmNarmCallback, symmetrical_cb, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("180 deg. rotate"))); n++;
  XtSetArg(wargs[n], XmNset, (symmetrical==2)?True:False); n++;
  symmetrical3 = XmCreateToggleButton(settings_radio2, "ToggleSimple23", wargs, n);
  XtManageChild(symmetrical3);
  XmStringFree(s[0]);
  XtAddCallback(symmetrical3, XmNarmCallback, symmetrical_cb, (XtPointer) 2);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_radio2); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  separator2 = XtCreateManagedWidget("separator2", xmSeparatorWidgetClass, settings_form, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, separator2); n++;
  settings_radio = XmCreateRadioBox(settings_form, "RadioBox1", wargs, n);
  XtManageChild(settings_radio);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Very easy"))); n++;
  XtSetArg(wargs[n], XmNset, (level==0)?True:False); n++;
  level1 = XmCreateToggleButton(settings_radio, "ToggleSimple1", wargs, n);
  XtManageChild(level1);
  XmStringFree(s[0]);
  XtAddCallback(level1, XmNarmCallback, level_cb, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Easy"))); n++;
  XtSetArg(wargs[n], XmNset, (level==1)?True:False); n++;
  level2 = XmCreateToggleButton(settings_radio, "ToggleSimple2", wargs, n);
  XtManageChild(level2);
  XmStringFree(s[0]);
  XtAddCallback(level2, XmNarmCallback, level_cb, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Medium (SSC)"))); n++;
  XtSetArg(wargs[n], XmNset, (level==2)?True:False); n++;
  level3 = XmCreateToggleButton(settings_radio, "ToggleSimple3", wargs, n);
  XtManageChild(level3);
  XmStringFree(s[0]);
  XtAddCallback(level3, XmNarmCallback, level_cb, (XtPointer) 2);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Hard (SSC+DS)"))); n++;
  XtSetArg(wargs[n], XmNset, (level==3)?True:False); n++;
  level4 = XmCreateToggleButton(settings_radio, "ToggleSimple4", wargs, n);
  XtManageChild(level4);
  XmStringFree(s[0]);
  XtAddCallback(level4, XmNarmCallback, level_cb, (XtPointer) 3);
  
  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Difficult (SSC+DS+Brute Force)"))); n++;
  XtSetArg(wargs[n], XmNset, (level==4)?True:False); n++;
  level5 = XmCreateToggleButton(settings_radio, "ToggleSimple5", wargs, n);
  XtManageChild(level5);
  XmStringFree(s[0]);
  XtAddCallback(level5, XmNarmCallback, level_cb, (XtPointer) 4);
  
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_radio); n++;
  XtSetArg(wargs[n], XmNtopOffset, 10); n++;
  XtSetArg(wargs[n], XmNbottomOffset, 10); n++;
  settings_ok = XtCreateManagedWidget("OK", xmPushButtonWidgetClass, settings_form, wargs, n);
  XtAddCallback(settings_ok, XmNactivateCallback, settingsok_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_ok); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_ok); n++;
  settings_cancel = XtCreateManagedWidget("Cancel", xmPushButtonWidgetClass, settings_form, wargs, n);
  XtAddCallback(settings_cancel, XmNactivateCallback, settingscancel_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XtNtitle, "PostScript File Selection"); n++;
  main_printdialog = XtCreateApplicationShell("main_printdialog", topLevelShellWidgetClass, wargs, n);

  n=0;
  printer_fileform = XtCreateManagedWidget("form", xmFormWidgetClass, main_printdialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNpattern, (stmp=XmStringCreateSimple("*.ps"))); n++;
  printer_fileselection = XtCreateManagedWidget("printer_fileselection", xmFileSelectionBoxWidgetClass, printer_fileform, wargs, n);
  XmStringFree(stmp);

  XtAddCallback(printer_fileselection, XmNokCallback, printer_fileopen_cb, (XtPointer)NULL);
  XtAddCallback(printer_fileselection, XmNcancelCallback, printer_fileclose_cb, (XtPointer)NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  papersize_label = XtCreateManagedWidget("Select paper size", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 0); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  papersize = XmCreateOptionMenu(printer_fileform, "papersize", wargs, n);

  n=0;
  papersize_pane = XmCreatePulldownMenu(papersize, "pulldown", wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("A4"))); n++;
  A4_w = XtCreateManagedWidget("A4_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(A4_w, XmNactivateCallback, printer_papersize_cb, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("Letter"))); n++;
  Letter_w = XtCreateManagedWidget("letter_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(Letter_w, XmNactivateCallback, printer_papersize_cb, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNsubMenuId, papersize_pane); n++;
  XtSetValues(papersize, wargs, n);

  XtManageChild(papersize);

  n=0;
  XtSetArg(wargs[n], XmNmenuHistory, letter?Letter_w:A4_w); n++;
  XtSetValues(papersize, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  printer_explain_label = XtCreateManagedWidget("Print explanation", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, printer_explain_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  printer_explain_check = XmCreateSimpleCheckBox(printer_fileform, "printer_explain_check", wargs, n);
  XtManageChild(printer_explain_check);

  n=0;
#if XmVERSION==1
  XtSetArg(wargs[n], XmNindicatorOn, True); n++;
  XtSetArg(wargs[n], XmNset, print_explain?True:False); n++;
#else
  XtSetArg(wargs[n], XmNindicatorOn, XmINDICATOR_CHECK_BOX); n++;
  XtSetArg(wargs[n], XmNset, print_explain?XmSET:XmUNSET); n++;
#endif
  XtSetArg(wargs[n], XmNindicatorSize, 20); n++;
  printer_explain_select = XmCreateToggleButton(printer_explain_check, "", wargs, n);
  XtManageChild(printer_explain_select);
  XmFileSelectionDoSearch(printer_fileselection, NULL);
  XtManageChild(main_wnd);
}

static void sudokumain()
{
  XEvent event;
  int status;

AGAIN:
  status = XtPending();
  if(status)
  {
AGAIN2:
    XtNextEvent(&event);
    XtDispatchEvent(&event);
    goto AGAIN;
  }
  if(running)
  {
    sudoku_run();
    goto AGAIN;
  }
  goto AGAIN2;
}  

main(int argc, char **argv)
{

  local_srandom(time(NULL));

  toplevel=XtInitialize("Sudoku", "Sudoku", NULL, 0, &argc, argv);

  XtGetApplicationResources(toplevel, &applicationdata,
                            resources, XtNumber(resources), NULL, 0);

  main_wnd_init();

  running = 0;

  XtRealizeWidget(toplevel);

  settitle("Sudoku generator");

  sudokumain();
}
